package com.netifera.platform.internal.model.sync;

import com.netifera.platform.api.dispatcher.IMessenger;
import com.netifera.platform.api.dispatcher.MessengerException;
import com.netifera.platform.api.events.IEvent;
import com.netifera.platform.api.events.IEventHandler;
import com.netifera.platform.api.log.ILogger;
import com.netifera.platform.internal.model.Workspace;
import com.netifera.platform.model.IUpdateRecord;
import com.netifera.platform.model.ModelUpdate;
/*
 * Handles synchronization for a single connection to the probe.
 */
public class ChannelSynchronizer extends Thread {
	final private ModelSynchronizer synchronizer;
	final private Workspace workspace;
	final private IMessenger messenger;
	final private ILogger logger;
	private long currentUpdateIndex;
	private long nextSendIndex;
	private volatile boolean done;
	
	ChannelSynchronizer(ModelSynchronizer synchronizer, IMessenger messenger, ILogger logger, long startIndex) {
		this.synchronizer = synchronizer;
		this.messenger = messenger;
		this.logger = logger;
		this.done = false;
		this.nextSendIndex = startIndex;
		this.workspace = synchronizer.getWorkspace();
		workspace.addEntityUpdateListener(new IEventHandler() {
			public void handleEvent(IEvent event) {
				event();	
			}
		});
	}

	private synchronized void event() {
		logger.debug("Event! update index = " + synchronizer.getWorkspace().getCurrentUpdateIndex());
		currentUpdateIndex = synchronizer.getWorkspace().getCurrentUpdateIndex();
		notify();
	}
	
	public void kill() {
		done = true;
		notify();
	}
	
	public void run() {
		currentUpdateIndex = synchronizer.getWorkspace().getCurrentUpdateIndex();

		logger.debug("Starting Channel Synchronizer");
		while(!done) {
			try {
				sendUpdates();
			} catch (InterruptedException e) {
				interrupt();
				done = true;
			}
		}
		synchronizer.removeChannelSynchronizer(messenger);
	}
	
	private void sendUpdates() throws InterruptedException {
		synchronized(this) {
			logger.debug("CurrentUpdateIndex: "  + currentUpdateIndex + " nextSendIndex: " + nextSendIndex);
			while(!done && currentUpdateIndex <= nextSendIndex) {
				wait();
			}
		}
		while(!done && nextSendIndex < currentUpdateIndex) {
			IUpdateRecord update = workspace.getEntityByUpdateIndex(nextSendIndex);
			incrementSendIndex();
			if(update != null) {
				sendUpdate(update);
			}
		}
		
	}
	private synchronized void incrementSendIndex() {
		nextSendIndex++;
	}
	
	private void sendUpdate(IUpdateRecord update) {
		logger.debug("Sending update record (" + (nextSendIndex - 1)  + ") " + update);
		final ModelUpdate updateMessage = new ModelUpdate();
		updateMessage.addUpdate(update);
		try {
			messenger.emitMessage(updateMessage);
		} catch (MessengerException e) {
			interrupt();
			done = true;
		}
	}
}
