package com.netifera.platform.net.internal.daemon.sniffing;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

import com.netifera.platform.api.dispatcher.IMessenger;
import com.netifera.platform.api.dispatcher.MessengerException;
import com.netifera.platform.api.log.ILogManager;
import com.netifera.platform.api.log.ILogger;
import com.netifera.platform.api.probe.IProbeManagerService;
import com.netifera.platform.net.daemon.sniffing.ISniffingDaemon;
import com.netifera.platform.net.daemon.sniffing.ISniffingModule;
import com.netifera.platform.net.daemon.sniffing.model.CaptureFileEntity;
import com.netifera.platform.net.daemon.sniffing.model.ISniffingEntityFactory;
import com.netifera.platform.net.daemon.sniffing.model.NetworkInterfaceEntity;
import com.netifera.platform.net.daemon.sniffing.model.SniffingSessionEntity;
import com.netifera.platform.net.internal.daemon.probe.SniffingModuleOutput;
import com.netifera.platform.net.pcap.ICaptureInterface;
import com.netifera.platform.net.sniffing.ICaptureFileInterface;
import com.netifera.platform.net.sniffing.ISniffingEngineService;
import com.netifera.platform.net.sniffing.util.ICaptureFileProgress;

public class SniffingDaemon implements ISniffingDaemon, ISniffingModuleOutput {
	
	/*
	 * All sniffing modules which have been registered with the daemon.
	 */
	private final Set<ISniffingModule> modules;
	
	/*
	 * Sniffing modules which have been enabled with the enableModules()
	 * method.
	 */
	private final Set<EnabledSniffingModule> enabledModules;
	
	/*
	 * Network interfaces which have been enabled with the enableInterfaces()
	 * method.
	 */
	private final Set<ICaptureInterface> enabledInterfaces;
	
	private IMessenger openMessenger;

	/*
	 * DS controlled services
	 */
	private ISniffingEngineService sniffingEngine;
	private ISniffingEntityFactory sniffingEntityFactory;
	private IProbeManagerService probeManager;
	private ILogger logger;
	/*
	 * Flag which indicates when the sniffing daemon is currently running.
	 */
	private boolean running;
	
	private ICaptureFileInterface runningCaptureFile;
	
	/**
	 * SniffingDaemon constructor creates a TreeSet for storing sniffing modules
	 * in alphabetic order.
	 */
	public SniffingDaemon() {
		modules = new TreeSet<ISniffingModule>(new Comparator<ISniffingModule>() {

			public int compare(ISniffingModule module1,
					ISniffingModule module2) {
				return module1.getName().compareTo(module2.getName());
			}
			
		});
		
		enabledModules = new HashSet<EnabledSniffingModule>();
		enabledInterfaces = new HashSet<ICaptureInterface>();
		running = false;
		
	}
	
	
	
	public Set<ISniffingModule> getModules() {
		return Collections.unmodifiableSet(modules);
	}
	
	
	public Collection<ICaptureInterface> getInterfaces() {
		return sniffingEngine.getInterfaces();
	}
	
	public boolean isEnabled(ICaptureInterface iface) {
		return enabledInterfaces.contains(iface);
	}
	
	public boolean hasEnabledInterfaces() {
		return !enabledInterfaces.isEmpty();
	}
	
	public void setEnabled(ICaptureInterface iface, boolean enable) {
		if(running) {
			throw new IllegalStateException("Cannot configure interfaces while daemon is running");
		}
		
		if(!getInterfaces().contains(iface)) {
			throw new IllegalArgumentException("Unknown interface passed to setEnabled()");
		}
		
		if(!iface.captureAvailable()) {
			throw new IllegalArgumentException("Cannot enable unavailable interface");
		}
		
		if(enable) {
			enabledInterfaces.add(iface);
		} else {
			enabledInterfaces.remove(iface);
		}
	}
	
	public boolean isEnabled(ISniffingModule module) {
		return findEnabledModule(module) != null;
	}
	
	public void setEnabled(ISniffingModule module, boolean enable) {
		if(running) {
			throw new IllegalStateException("Cannot configure modules while daemon is running");
		}
		if(!getModules().contains(module)) {
			throw new IllegalArgumentException("Unknown module passed to setEnabled()");
		}
		if(enable) {
			if(findEnabledModule(module) == null) {
				enabledModules.add(new EnabledSniffingModule(sniffingEngine, module, this, logger));
			}
		} else {
			EnabledSniffingModule esm = findEnabledModule(module);
			if(esm != null) {
				enabledModules.remove(esm);
			}
		}
	}
	
	private EnabledSniffingModule findEnabledModule(ISniffingModule module) {
		for(EnabledSniffingModule m : enabledModules) {
			if(m.getModule().equals(module)) {
				return m;
			}
		}
		return null;
	}
	
	public void enableAllModules() {
		enableModules(getModules());
	}
	
	public synchronized void enableModules(Set<ISniffingModule> enabledModuleSet) {
		if(running) {
			throw new IllegalStateException("Cannot enable modules while daemon is running");
		}
		
		if(enabledModuleSet.isEmpty()) {
			throw new IllegalArgumentException("enableModules() called with empty set");
		}
		
		enabledModules.clear();
		for(ISniffingModule module : enabledModuleSet) {
			if(!modules.contains(module)) {
				throw new IllegalArgumentException("Unknown Sniffing Module requested in enableModules() " +
						module);
			}
			enabledModules.add(new EnabledSniffingModule(sniffingEngine, module, this, logger));
		}
	}
	
	public void enableAllInterfaces() {
		enableInterfaces(getInterfaces());
	}
	
	public synchronized void enableInterfaces(Collection<ICaptureInterface> interfaces) {
		if(running) {
			throw new IllegalStateException("Cannot enable interfaces while daemon is running");
		}
		
		if(interfaces.isEmpty()) {
			throw new IllegalArgumentException("enableInterfaces() called with empty collection");
		}
		enabledInterfaces.clear();
		for(ICaptureInterface iface : interfaces) {
			if(!iface.captureAvailable()) {
				continue;
			}
			if(!getInterfaces().contains(iface)) {
				throw new IllegalArgumentException("Unknown network interface requested in enableInterfaces() "
					+ iface);
			}
			enabledInterfaces.add(iface);
		}
	}
	
	public ICaptureFileInterface createCaptureFileInterface(String path) {
		return sniffingEngine.createCaptureFileInterface(path);
	}
	
	public synchronized void runCaptureFile(long spaceId, ICaptureFileInterface iface, final ICaptureFileProgress progress) {
		
		if(!iface.isValid()) {
			progress.error("Invalid interface " + iface.getName(), null);
			return;
		}
		final long realm = probeManager.getLocalProbe().getEntity().getId();
		final CaptureFileEntity captureFile = sniffingEntityFactory.createCaptureFile(realm, spaceId, iface.getPath());
		
		
		final Set<SniffingDaemonInterface> interfaces = new HashSet<SniffingDaemonInterface>();

		interfaces.add(new SniffingDaemonInterface(iface, captureFile.getId()));
		
		for(EnabledSniffingModule module : enabledModules) {
			module.start(interfaces, spaceId);
		}
		
		running = true;
		runningCaptureFile = iface;
		iface.process(new ICaptureFileProgress() {

			public void error(String message, Throwable e) {
				stop();
				progress.error(message, e);
			}

			public void finished() {
				runningCaptureFile = null;
				stop();
				progress.finished();
			}

			public boolean updateProgress(int percent, int count) {
				return progress.updateProgress(percent, count);
			}
		});
		
	}
	
	public void cancelCaptureFile() {
		if(runningCaptureFile != null) runningCaptureFile.cancelProcessing();
	}
	
	public synchronized void start(long spaceId) {
		if(running) {
			return;
		}
		
		final long realm = probeManager.getLocalProbe().getEntity().getId();
		final SniffingSessionEntity session = sniffingEntityFactory.createSniffingSession(realm, spaceId);
		final Set<SniffingDaemonInterface> interfaces = new HashSet<SniffingDaemonInterface>();
		for(ICaptureInterface iface : enabledInterfaces) {
			NetworkInterfaceEntity interfaceEntity = sniffingEntityFactory.createNetworkInterface(session.getId(), spaceId, iface.getName());
			interfaces.add(new SniffingDaemonInterface(iface, interfaceEntity.getId()));
		}
		for(EnabledSniffingModule module : enabledModules) {
			module.start(interfaces, spaceId);
		}
		running = true;
		
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.netifera.platform.api.daemon.IDaemon#stop()
	 */
	public synchronized void stop() {
		if(!running) {
			return;
		}
		for(EnabledSniffingModule module : enabledModules) {
			module.stop();
		}
		running = false;
	}
	
	public boolean isRunning() {
		return running;
	}
	
	public void setOpenMessenger(IMessenger messenger) {
		openMessenger = messenger;
	}
	
	public void printOutput(String message) {
		if(openMessenger == null) return;
		if (message.endsWith("\n")) {
			message = message.concat("\n");
		}
		try {
			openMessenger.emitMessage(new SniffingModuleOutput(message));
		} catch (MessengerException e) {
			openMessenger = null;
		}
	}
	
	/*
	 * link to the component
	 */
	void registerModule(ISniffingModule module) {
		modules.add(module);
	}
	
	void unregisterModule(ISniffingModule module) {
		modules.remove(module);
	}
	
	void setSniffingEngine(ISniffingEngineService engine) {
		sniffingEngine = engine;
	}
	
	void setEntityFactory(ISniffingEntityFactory factory) {
		sniffingEntityFactory = factory;
	}
	
	void setProbeManager(IProbeManagerService manager) {
		probeManager = manager;
	}
	
	void setLogManager(ILogManager logManager) {
		logger = logManager.getLogger("Sniffing Daemon");
	}
	
	
}