package com.netifera.platform.net.internal.daemon.sniffing;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.service.component.ComponentContext;

import com.netifera.platform.api.dispatcher.DispatchException;
import com.netifera.platform.api.dispatcher.DispatchMismatchException;
import com.netifera.platform.api.dispatcher.IMessageDispatcher;
import com.netifera.platform.api.dispatcher.IMessageDispatcherService;
import com.netifera.platform.api.dispatcher.IMessageHandler;
import com.netifera.platform.api.dispatcher.IMessenger;
import com.netifera.platform.api.dispatcher.IProbeMessage;
import com.netifera.platform.api.dispatcher.MessengerException;
import com.netifera.platform.api.log.ILogManager;
import com.netifera.platform.api.log.ILogger;
import com.netifera.platform.api.probe.IProbeManagerService;
import com.netifera.platform.net.daemon.sniffing.ISniffingModule;
import com.netifera.platform.net.daemon.sniffing.model.ISniffingEntityFactory;
import com.netifera.platform.net.internal.daemon.probe.CancelCaptureFile;
import com.netifera.platform.net.internal.daemon.probe.CaptureFileValid;
import com.netifera.platform.net.internal.daemon.probe.InterfaceRecord;
import com.netifera.platform.net.internal.daemon.probe.ModuleRecord;
import com.netifera.platform.net.internal.daemon.probe.RequestInterfaceInformation;
import com.netifera.platform.net.internal.daemon.probe.RequestModuleInformation;
import com.netifera.platform.net.internal.daemon.probe.RunCaptureFile;
import com.netifera.platform.net.internal.daemon.probe.SetInterfaceEnableState;
import com.netifera.platform.net.internal.daemon.probe.SetModuleEnableState;
import com.netifera.platform.net.internal.daemon.probe.SniffingDaemonStatus;
import com.netifera.platform.net.internal.daemon.probe.StartSniffingDaemon;
import com.netifera.platform.net.internal.daemon.probe.StopSniffingDaemon;
import com.netifera.platform.net.pcap.ICaptureInterface;
import com.netifera.platform.net.sniffing.ICaptureFileInterface;
import com.netifera.platform.net.sniffing.ISniffingEngineService;

public class SniffingDaemonComponent {
	private final SniffingDaemon daemon = new SniffingDaemon();
	private final Map<String, ISniffingModule> moduleByName = new HashMap<String, ISniffingModule>();
	private ILogger logger;
	private Thread captureFileThread;
	
	
	void printOutput(String message) {
		
	}
	private void requestInterfaceInformation(IMessenger messenger, RequestInterfaceInformation msg) throws MessengerException {
		final List<InterfaceRecord> interfaces = new ArrayList<InterfaceRecord>();
		for(ICaptureInterface iface : daemon.getInterfaces()) {
			interfaces.add(new InterfaceRecord(iface.getName(), iface.toString(), iface.captureAvailable(), daemon.isEnabled(iface)));
		}
		messenger.emitMessage(msg.createResponse(interfaces));
	}
	
	private void requestModuleInformation(IMessenger messenger, RequestModuleInformation msg) throws MessengerException {
		final List <ModuleRecord> modules = new ArrayList<ModuleRecord>();
		for(ISniffingModule mod : daemon.getModules()) {
			modules.add(new ModuleRecord(mod.getName(), daemon.isEnabled(mod)));
		}
		messenger.emitMessage(msg.createResponse(modules));
	}
	private void setInterfaceEnableState(IMessenger messenger, SetInterfaceEnableState msg) throws MessengerException {
		for(InterfaceRecord iface : msg.getInterfaceRecords()) {
			ICaptureInterface captureInterface = lookupInterfaceByName(iface.getName());
			if (captureInterface == null) {
				logger.warning("No capture interface found with name " + iface.getName());
			} else {
				daemon.setEnabled(captureInterface, iface.isEnabled());
			}
		}
		messenger.respondOk(msg);
	}
	
	private ICaptureInterface lookupInterfaceByName(String name) {
		for(ICaptureInterface iface : daemon.getInterfaces()) {
			if(iface.getName().equals(name))
				return iface;
		}
		return null;
	}
	private void setModuleEnableState(IMessenger messenger, SetModuleEnableState msg) throws MessengerException {
		for(ModuleRecord module : msg.getModuleRecords()) {
			ISniffingModule sniffingModule = moduleByName.get(module.getName());
			if (sniffingModule == null) {
				logger.warning("No sniffing module found with name " + module.getName());
			} else {
				daemon.setEnabled(sniffingModule, module.isEnabled());
			}
		}
		messenger.respondOk(msg);
	}
	
	private void startSniffingDaemon(IMessenger messenger, StartSniffingDaemon msg) throws MessengerException {
		daemon.setOpenMessenger(messenger);
		daemon.start(msg.getSpaceId());
		messenger.respondOk(msg);
	}
	
	private void stopSniffingDaemon(IMessenger messenger, StopSniffingDaemon msg) throws MessengerException {
		daemon.stop();
		messenger.respondOk(msg);
	}
	
	private void captureFileValid(IMessenger messenger, CaptureFileValid msg) throws MessengerException {
		ICaptureFileInterface iface = daemon.createCaptureFileInterface(msg.getPath());
		messenger.emitMessage(msg.createResponse(iface.isValid(), iface.getErrorMessage()));
		iface.dispose();
	}
	
	private void sniffingDaemonStatus(IMessenger messenger, SniffingDaemonStatus msg) throws MessengerException {
		messenger.emitMessage(msg.createResponse(daemon.isRunning()));
	}
	
	private void runCaptureFile(IMessenger messenger, RunCaptureFile msg) throws MessengerException {
		ICaptureFileInterface iface = daemon.createCaptureFileInterface(msg.getPath());
		if(!iface.isValid()) {
			messenger.respondError(msg, iface.getErrorMessage());
			return;
		} else {
			messenger.respondOk(msg);
		}
		daemon.setOpenMessenger(messenger);
		CaptureFileRunnable runnable = new CaptureFileRunnable(messenger, iface, daemon, msg.getSpaceId());
		captureFileThread = new Thread(runnable);
		captureFileThread.start();
	}
	
	private void cancelCaptureFile(IMessenger messenger, CancelCaptureFile msg) throws MessengerException {
		daemon.cancelCaptureFile();
		messenger.respondOk(msg);
	}
	
	private void registerHandlers(IMessageDispatcher dispatcher) {
		IMessageHandler handler = new IMessageHandler() {

			public void call(IMessenger messenger, IProbeMessage message)
					throws DispatchException {
				try {
					dispatch(messenger, message);
				} catch (MessengerException e) {
					logger.warning("Error sending message response " + e.getMessage());
				}
			}
		};
		
		dispatcher.registerMessageHandler(RequestInterfaceInformation.ID, handler);
		dispatcher.registerMessageHandler(RequestModuleInformation.ID, handler);
		dispatcher.registerMessageHandler(SetInterfaceEnableState.ID, handler);
		dispatcher.registerMessageHandler(SetModuleEnableState.ID, handler);
		dispatcher.registerMessageHandler(StartSniffingDaemon.ID, handler);
		dispatcher.registerMessageHandler(StopSniffingDaemon.ID, handler);
		dispatcher.registerMessageHandler(CaptureFileValid.ID, handler);
		dispatcher.registerMessageHandler(SniffingDaemonStatus.ID, handler);
		dispatcher.registerMessageHandler(RunCaptureFile.ID, handler);
		dispatcher.registerMessageHandler(CancelCaptureFile.ID, handler);
	}
	
	private void dispatch(IMessenger messenger, IProbeMessage message) throws DispatchMismatchException, MessengerException {
		if(message instanceof RequestInterfaceInformation) {
			requestInterfaceInformation(messenger, (RequestInterfaceInformation) message);
		} else if(message instanceof RequestModuleInformation) {
			requestModuleInformation(messenger, (RequestModuleInformation) message);
		} else if(message instanceof SetInterfaceEnableState) {
			setInterfaceEnableState(messenger, (SetInterfaceEnableState) message);
		} else if(message instanceof SetModuleEnableState) {
			setModuleEnableState(messenger, (SetModuleEnableState) message);
		} else if(message instanceof StartSniffingDaemon) {
			startSniffingDaemon(messenger, (StartSniffingDaemon) message);
		} else if(message instanceof StopSniffingDaemon) {
			stopSniffingDaemon(messenger, (StopSniffingDaemon) message);
		} else if(message instanceof CaptureFileValid) {
			captureFileValid(messenger, (CaptureFileValid) message);
		} else if(message instanceof SniffingDaemonStatus) {
			sniffingDaemonStatus(messenger, (SniffingDaemonStatus) message);
		} else if(message instanceof RunCaptureFile) {
			runCaptureFile(messenger, (RunCaptureFile) message);
		} else if(message instanceof CancelCaptureFile) {
			cancelCaptureFile(messenger, (CancelCaptureFile) message);
		}else {
			throw new DispatchMismatchException(message);
		}
	}
	
	private boolean isActivated = false;

	protected void activate(ComponentContext ctx) {
		synchronized(daemon) {
			daemon.enableAllInterfaces();
			daemon.enableAllModules();
			isActivated = true;
		}
	}
	
	protected void deactivate(ComponentContext ctx) {
	}
	
	
	/*
	 * OSGi DS binding functions
	 */
	protected void registerModule(ISniffingModule module) {
		synchronized(daemon) {
			daemon.registerModule(module);
			if(isActivated)
				daemon.setEnabled(module, true);
		}
		
		moduleByName.put(module.getName(), module);		
	}
	
	protected void unregisterModule(ISniffingModule module) {
		daemon.unregisterModule(module);
	}
	
	protected void setSniffingEngine(ISniffingEngineService engine) {
		daemon.setSniffingEngine(engine);
	}
	
	protected void unsetSniffingEngine(ISniffingEngineService engine) {
	}


	protected void setEntityFactory(ISniffingEntityFactory factory) {
		daemon.setEntityFactory(factory);
	}
	
	protected void unsetEntityFactory(ISniffingEntityFactory factory) {
	}
	
	protected void setProbeManager(IProbeManagerService manager) {
		daemon.setProbeManager(manager);
	}
	
	protected void unsetProbeManager(IProbeManagerService manager) {
	}
	
	protected void setDispatcher(IMessageDispatcherService dispatcherService) {
		registerHandlers(dispatcherService.getServerDispatcher());
	}

	protected void unsetDispatcher(IMessageDispatcherService dispatcherService) {
		
	}
	
	protected void setLogManager(ILogManager logManager) {
		logger = logManager.getLogger("Sniffing Daemon");
		daemon.setLogManager(logManager);
	}
	
	protected void unsetLogManager(ILogManager logManager) {
		
	}


}
