package com.netifera.platform.net.internal.pcap;

import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;

import org.osgi.service.component.ComponentContext;

import com.netifera.platform.api.log.ILogManager;
import com.netifera.platform.api.log.ILogger;
import com.netifera.platform.api.system.ISystemService;
import com.netifera.platform.net.internal.pcap.linux.LinuxPacketCapture;
import com.netifera.platform.net.internal.pcap.osx.OsxPacketCapture;
import com.netifera.platform.net.pcap.ICaptureInterface;
import com.netifera.platform.net.pcap.IPacketCapture;
import com.netifera.platform.net.pcap.IPacketCaptureFactoryService;
import com.netifera.platform.net.pcap.IPacketHandler;

public class PacketCaptureFactoryService implements
		IPacketCaptureFactoryService {

	private ISystemService system;
	private ILogger logger;

	private final Set<ICaptureInterface> interfaces = new HashSet<ICaptureInterface>();
	
	public IPacketCapture create(ICaptureInterface iface, int snaplen, boolean promiscuous,
			int timeout, IPacketHandler packetHandler) {
		
		if(!(iface instanceof PhysicalInterface)) {
			throw new IllegalArgumentException("Interface argument must be a physical interface");	
		}
		
		PhysicalInterface phys = (PhysicalInterface) iface;
		
		PacketCapture pcap = new PacketCapture(phys.getInterface(), snaplen, promiscuous, timeout, packetHandler);
		
		INativePacketCapture nativePcap = createNative(pcap);
		pcap.setNativeCapture(nativePcap);
		
		return pcap;
	}
	
	
	public Collection<ICaptureInterface> getInterfaces() {
		synchronized(interfaces) {
			return new HashSet<ICaptureInterface>(interfaces);
		}
	}
	
	public Collection<ICaptureInterface> getCurrentInterfaces() {
		try {
			initializeInterfaces();
		} catch (SocketException e) {
			return Collections.emptySet();
		}
		return getInterfaces();
	}
	
	private INativePacketCapture createNative(IPacketCaptureInternal pcap) {
		switch(system.getOS()) {
		case OS_OSX:
			return new OsxPacketCapture(system, pcap, logger);
			
		case OS_LINUX:
			return new LinuxPacketCapture(system, pcap, logger);
			
		default:
			throw new IllegalStateException("No native pcap implementation for current OS");
		}
	
	}

	protected void setSystemService(ISystemService system) {
		this.system = system;
	}
	
	protected void unsetSystemService(ISystemService system) {
		this.system = null;
	}
	
	protected void activate(ComponentContext ctx) {
		try {
			initializeInterfaces();
		} catch (SocketException e) {
			throw new RuntimeException("SocketException initializing interfaces", e);
		}
	}
	
	protected void deactivate(ComponentContext ctx) {
		interfaces.clear();
	}
	
	protected void setLogManager(ILogManager logManager) {
		logger = logManager.getLogger("PacketCapture");
	}
	
	protected void unsetLogManager(ILogManager logManager) {
		logger = null;
	}
	
	private void initializeInterfaces() throws SocketException {
		Enumeration<NetworkInterface> networkInterfaces = NetworkInterface.getNetworkInterfaces();
		
		if(networkInterfaces == null) {
			return;
		}
		
		synchronized(interfaces) {
			interfaces.clear();
		
			while(networkInterfaces.hasMoreElements()) {
				NetworkInterface iface = networkInterfaces.nextElement();
				PacketCapture pcap = new PacketCapture(iface);
				INativePacketCapture nativePcap = createNative(pcap);
				pcap.setNativeCapture(nativePcap);
				interfaces.add(new PhysicalInterface(iface, pcap.testOpen()));
			}
		}
			
		
	}
}
